use std::{
    cmp::{max, min},
    ops::RangeInclusive,
};

use crate::{
    constants::{
        LEVEL_HEIGHT, LEVEL_WIDTH, ROOM_HEIGHT, ROOM_MIN_HEIGHT, ROOM_MIN_WIDTH, ROOM_WIDTH,
    },
    level::StructureElement,
};
use rand::rngs::ThreadRng;
use rand::Rng;

#[derive(PartialEq, Copy, Clone, Eq, Hash, Debug)]
pub enum RoomType {
    Start,
    End,
    StairUp,
    StairDown,
    BasicRoom,
    ArtifactRoom,
    MonsterRoom,
    EmptyRoom,
}

#[derive(Copy, Clone, Debug)]
pub struct Connection {
    pub start_pos: (usize, usize),
    pub end_pos: (usize, usize),
}

impl Connection {
    pub fn render(
        &self,
        rng: &mut ThreadRng,
        tgt: &mut [[StructureElement; LEVEL_HEIGHT]; LEVEL_WIDTH],
    ) {
        fn calc_range(a: usize, b: usize) -> RangeInclusive<usize> {
            if a > b {
                b..=a
            } else {
                a..=b
            }
        }
        // the tuples are (col, row)
        let abs_d_col =
            max(self.start_pos.0, self.end_pos.0) - min(self.start_pos.0, self.end_pos.0);

        let abs_d_row =
            max(self.start_pos.1, self.end_pos.1) - min(self.start_pos.1, self.end_pos.1);

        if abs_d_col == 0 {
            for row in calc_range(self.end_pos.1, self.start_pos.1) {
                tgt[self.end_pos.0][row] = StructureElement::Floor;
            }
        } else if abs_d_row == 0 {
            for col in calc_range(self.end_pos.0, self.start_pos.0) {
                tgt[col][self.end_pos.1] = StructureElement::Floor;
            }
        } else {
            let d_row = self.end_pos.1 as i32 - self.start_pos.1 as i32;
            if d_row > 0 {
                // more up/down
                let d_row = self.end_pos.1 - self.start_pos.1;
                let vert_offset = if d_row > 3 {
                    rng.gen_range(1..=d_row - 1)
                } else {
                    d_row / 2
                };
                for r in self.start_pos.1..=(self.start_pos.1 + vert_offset) {
                    tgt[self.start_pos.0][r] = StructureElement::Floor;
                }
                for c in calc_range(self.end_pos.0, self.start_pos.0) {
                    tgt[c][self.start_pos.1 + vert_offset] = StructureElement::Floor;
                }
                for r in (self.start_pos.1 + vert_offset)..=self.end_pos.1 {
                    tgt[self.end_pos.0][r] = StructureElement::Floor;
                }
            } else {
                // more left/right
                let d_col = self.end_pos.0 - self.start_pos.0;
                let horizont_offset = if d_col > 3 {
                    rng.gen_range(1..=d_col - 1)
                } else {
                    d_col / 2
                };
                for tgt_col in tgt
                    .iter_mut()
                    .skip(self.start_pos.0)
                    .take(horizont_offset + 1)
                {
                    tgt_col[self.start_pos.1] = StructureElement::Floor;
                }
                for r in calc_range(self.end_pos.1, self.start_pos.1) {
                    tgt[self.start_pos.0 + horizont_offset][r] = StructureElement::Floor;
                }
                for tgt_col in tgt
                    .iter_mut()
                    .take(self.end_pos.0 + 1)
                    .skip(self.start_pos.0 + horizont_offset)
                {
                    tgt_col[self.end_pos.1] = StructureElement::Floor;
                }
            }
        }
    }
}

#[derive(Copy, Clone, Debug)]
pub struct Room {
    pub kind: RoomType,
    pub offset_x: usize,
    pub offset_y: usize,
    pub width: usize,
    pub height: usize,
    pub special: (usize, usize),
    pub connection_down: Option<Connection>,
    pub connection_right: Option<Connection>,
}

impl Room {
    pub fn new(rng: &mut ThreadRng) -> Self {
        let width = rng.gen_range(ROOM_MIN_WIDTH..ROOM_WIDTH);
        let height = rng.gen_range(ROOM_MIN_HEIGHT..ROOM_HEIGHT);
        let offset_x = rng.gen_range(0..(ROOM_WIDTH - width));
        let offset_y = rng.gen_range(0..(ROOM_HEIGHT - height));
        let sx = offset_x + rng.gen_range(1..width - 1);
        let sy = offset_y + rng.gen_range(1..height - 1);

        Self {
            kind: RoomType::EmptyRoom,
            offset_x,
            offset_y,
            width,
            height,
            special: (sx, sy),
            connection_down: None,
            connection_right: None,
        }
    }

    // pub fn get_x_range(&self) -> Range<usize> {
    //     self.offset_x..self.offset_x + self.width
    // }

    // pub fn get_y_range(&self) -> Range<usize> {
    //     self.offset_y..self.offset_y + self.height
    // }

    pub fn render(
        &self,
        tgt: &mut [[StructureElement; LEVEL_HEIGHT]; LEVEL_WIDTH],
        col: usize,
        row: usize,
    ) -> (usize, usize) {
        let top = 1 + row * ROOM_HEIGHT;
        let left = 1 + col * ROOM_WIDTH;
        for col in tgt.iter_mut().skip(left).take(ROOM_WIDTH) {
            //left..left + ROOM_WIDTH {
            for row_element in col.iter_mut().skip(top).take(ROOM_HEIGHT) {
                // top..top + ROOM_HEIGHT {
                *row_element = StructureElement::Wall;
            }
        }
        if self.kind == RoomType::EmptyRoom {
            return (0, 0);
        }
        // render floor elements
        for x in 0..self.width {
            for y in 0..self.height {
                tgt[left + self.offset_x + x][top + self.offset_y + y] = StructureElement::Floor;
            }
        }

        match self.kind {
            RoomType::Start => {
                tgt[left + self.special.0][top + self.special.1] = StructureElement::Start;
            }
            RoomType::End => {
                tgt[left + self.special.0][top + self.special.1] = StructureElement::End;
            }
            RoomType::StairUp => {
                tgt[left + self.special.0][top + self.special.1] = StructureElement::StairUp;
            }
            RoomType::StairDown => {
                tgt[left + self.special.0][top + self.special.1] = StructureElement::StairDown;
            }
            _ => {}
        };
        (left + self.special.0, top + self.special.1)
    }
}

#[test]
fn test_room_creation() {
    let mut rng = rand::thread_rng();
    let room = Room::new(&mut rng);
    assert_eq!(room.kind, RoomType::EmptyRoom);
    for _ in 0..1000 {
        let room = Room::new(&mut rng);
        assert!((3..=ROOM_WIDTH).contains(&room.width));
    }
}

#[cfg(test)]
fn all_wall(
    tgt: &mut [[StructureElement; LEVEL_HEIGHT]; LEVEL_WIDTH],
    col: usize,
    row: usize,
) -> bool {
    let top = 1 + row * ROOM_HEIGHT;
    let left = 1 + col * ROOM_WIDTH;
    for x in left..left + ROOM_WIDTH {
        for y in top..top + ROOM_HEIGHT {
            if tgt[x][y] != StructureElement::Wall {
                return false;
            }
        }
    }
    true
}

#[cfg(test)]
fn has_structure_element(
    tgt: &mut [[StructureElement; LEVEL_HEIGHT]; LEVEL_WIDTH],
    col: usize,
    row: usize,
    element: StructureElement,
) -> bool {
    let top = 1 + row * ROOM_HEIGHT;
    let left = 1 + col * ROOM_WIDTH;
    for x in left..left + ROOM_WIDTH {
        for y in top..top + ROOM_HEIGHT {
            if tgt[x][y] == element {
                return true;
            }
        }
    }
    false
}

#[test]
fn test_room_render_empty() {
    let mut rng = rand::thread_rng();
    let room = Room::new(&mut rng);
    assert_eq!(room.kind, RoomType::EmptyRoom);
    let mut structure = [[StructureElement::Floor; LEVEL_HEIGHT]; LEVEL_WIDTH];

    room.render(&mut structure, 0, 0);
    assert!(all_wall(&mut structure, 0, 0));
}

#[test]
fn test_room_render_basic() {
    let mut rng = rand::thread_rng();
    let mut room = Room::new(&mut rng);
    room.kind = RoomType::BasicRoom;
    let mut structure = [[StructureElement::Floor; LEVEL_HEIGHT]; LEVEL_WIDTH];

    room.render(&mut structure, 1, 3);
    assert!(!all_wall(&mut structure, 1, 3));
}

#[test]
fn test_room_render_start() {
    let mut rng = rand::thread_rng();
    let mut room = Room::new(&mut rng);
    room.kind = RoomType::Start;
    let mut structure = [[StructureElement::Floor; LEVEL_HEIGHT]; LEVEL_WIDTH];

    room.render(&mut structure, 1, 3);
    assert!(!all_wall(&mut structure, 1, 3));
    assert!(has_structure_element(
        &mut structure,
        1,
        3,
        StructureElement::Start
    ));
}
